IMPLEMENTATION MODULE BMP;

IMPORT IO;
IMPORT FIO;
IMPORT Graph;
FROM Storage IMPORT ALLOCATE,DEALLOCATE,Available;

(*Declaracion de la cabezera del archivo BMP para despues leerlo*)

TYPE

   CABECERABMP=RECORD
                   bfType:INTEGER; (*Simplemente "BM"*)
 		   bfSize:LONGINT; (*Tamao del fichero*)
 		   bfReserved1:INTEGER; (*Siempre 0*)
 		   bfReserved2:INTEGER; (*Siempre 0*)
 		   bfOffbits:LONGINT; (*desplazamiento en bytes desde esta cabezera
 		   			al inicio de la imagen*)

		   (*A partir de aqui lo que tenemos son datos sobre la estructura de la imagen*)
	           biSize:LONGINT; (*Bytes que ocupa el registro*)
 		   biWidth:LONGINT; (*Anchura en pixels*)
 		   biHeight:LONGINT; (*Altura en pixels*)
 		   biPlanes:INTEGER; (*Siempre 1*)
 		   biBitCount:INTEGER; (*Bits por pixel, intentaremos que sea 24*)
 		   biCompression:LONGINT; (*Siempre 0*)
 		   biSizeImage:LONGINT; (*Tamao en bytes de la imagen*)
 		   biXpelsPerMeter:LONGINT; (*Siempre 0*)
 		   biYpelsPerMeter:LONGINT; (*Siempre 0*)
 		   biClrUsed:LONGINT; (*Siempre 0*)
 		   biClrImportant:LONGINT; (*Siempre 0*)
  		END;

     CUERPOBMP=POINTER TO NODOBYTE;
     NODOBYTE=RECORD
                   byte:BYTE;
                   sig:CUERPOBMP;
              END;
     NODOBMP=RECORD
	          head:CABECERABMP;
	          body:CUERPOBMP;
	     END;
     TBMP=POINTER TO NODOBMP;


PROCEDURE Inicializa(VAR dibujo:TBMP):INTEGER;


BEGIN
    dibujo:=NIL;
    RETURN 0;

END Inicializa;

PROCEDURE Destruir(VAR dibujo:TBMP):INTEGER;

   PROCEDURE DestruyeCuerpo(VAR body:CUERPOBMP);
   VAR
     temp:CUERPOBMP;
   BEGIN
       WHILE body<>NIL DO
             temp:=body;
             body:=body^.sig;
             DISPOSE(temp);
       END;
   END DestruyeCuerpo;
BEGIN
    IF dibujo<>NIL THEN
       DestruyeCuerpo(dibujo^.body);
       DISPOSE(dibujo);
       dibujo:=NIL;
    END;
    RETURN 0
END Destruir;

PROCEDURE CargaArchivo(VAR dibujo:TBMP;nombre:ARRAY OF CHAR):INTEGER;

  PROCEDURE InsertaCuerpo(VAR lista:CUERPOBMP; byteleido:BYTE);

  VAR
    temp:CUERPOBMP;

  BEGIN
      NEW(temp);
      temp^.byte:=byteleido;
      temp^.sig:=lista;
      lista:=temp;

  END InsertaCuerpo;


VAR
  archivo:FIO.File;
  leidos,j:CARDINAL;
  control:INTEGER;
  byteleido:BYTE;
  i,totalpixels,bytesacargar:LONGINT;


BEGIN
    (*Comprobamos que exista el archivo*)
    IF NOT FIO.Exists(nombre) THEN
       RETURN -1;
    END;

    (*El archivo existe,procedemos a abrirlo*)
    archivo:=FIO.Open(nombre);

    IF NOT (dibujo=NIL) THEN (*La variable esta llena*)
       control:=Destruir(dibujo);
    END;	
    NEW(dibujo);
    dibujo^.body:=NIL;
    leidos:=FIO.RdBin(archivo,dibujo^.head,SIZE(CABECERABMP));

    (*Comprobamos que en la cabecera ponga que el archivo es BMP*)
    IF dibujo^.head.bfType<>19778 THEN
       RETURN -2
    END;

    (*Ahora cargamos el cuerpo del archivo en una estructura dinamica*)

    (*Posicionamos el manejador al principio de la zona de datos*)
    FIO.Seek(archivo,dibujo^.head.bfOffbits);

    totalpixels:=dibujo^.head.biWidth*dibujo^.head.biHeight;
    bytesacargar:=(dibujo^.head.bfSize-dibujo^.head.bfOffbits);

    (*IF (totalpixels MOD 2)=0 THEN
       bytesacargar:=totalpixels DIV 2
    ELSE
       bytesacargar:=totalpixels DIV 2 +1
    END;*)

    FOR i:=1 TO bytesacargar DO
        IF NOT Available(SIZE(NODOBYTE)) THEN (*Comprobamos que tenemos memoria*)
           RETURN -3
        END;
        leidos:=FIO.RdBin(archivo,byteleido,SIZE(BYTE));
        InsertaCuerpo(dibujo^.body,byteleido);
    END;


    FIO.Close(archivo);
    RETURN 0;
END CargaArchivo;



PROCEDURE ConvertColor(color:CARDINAL):CARDINAL;


BEGIN
    CASE color OF
         |1:RETURN 12;
         |2:RETURN 10;
         |3:RETURN 14;
         |4:RETURN 9;
         |5:RETURN 13;
         |6:RETURN 11;
         |7:RETURN 7;
         |8:RETURN 8;
         |9:RETURN 4;
         |10:RETURN 2;
         |11:RETURN 6;
         |12:RETURN 1;
         |13:RETURN 5;
         |14:RETURN 3;

    ELSE
       RETURN color
    END;

END ConvertColor;

PROCEDURE VisBmpMask(VAR dibujo:TBMP;x,y,mask:CARDINAL):INTEGER;

VAR
  i,j,altura,anchura:LONGINT;
  punt:CUERPOBMP;
  bytepart:BOOLEAN;
  color:CARDINAL;

BEGIN
    IF dibujo=NIL THEN (*Si el dibujo no ha sido cargado*)
       RETURN -1
    END;
    bytepart:=FALSE;
    anchura:=dibujo^.head.biWidth;
    altura:=dibujo^.head.biHeight;
    punt:=dibujo^.body;


    FOR i:=1 TO altura  DO
        FOR j:=anchura TO 1 BY (-1) DO
            IF punt=NIL THEN
               RETURN -2
            END;
            IF bytepart THEN
               color:=(CARDINAL(punt^.byte) DIV 16);
               color:=ConvertColor(color);
               IF (color<>mask) THEN
               	  Graph.Plot(CARDINAL(j)+x,CARDINAL(i)+y,color);
               END;
               bytepart:=FALSE;
               punt:=punt^.sig;
            ELSE
               color:=(CARDINAL(punt^.byte) MOD 16);
               color:=ConvertColor(color);
               IF (color<>mask) THEN
                  Graph.Plot(CARDINAL(j)+x,CARDINAL(i)+y,color);
               END;
               bytepart:=TRUE; (*Para la siguiente vuelta*)
            END;
        END;
    END;
    RETURN 0
END VisBmpMask;

PROCEDURE VisBmp(VAR dibujo:TBMP;x,y:CARDINAL):INTEGER;

BEGIN
    RETURN VisBmpMask(dibujo,x,y,17);

END VisBmp;

PROCEDURE Anchura(VAR dibujo:TBMP):CARDINAL;

BEGIN
    IF dibujo=NIL THEN
       RETURN 0
    ELSE
       RETURN CARDINAL(dibujo^.head.biWidth);
    END;
END Anchura;


PROCEDURE Altura(VAR dibujo:TBMP):CARDINAL;

BEGIN
    IF dibujo=NIL THEN
       RETURN 0;
    ELSE
        RETURN CARDINAL(dibujo^.head.biHeight);
    END;

END Altura;

END BMP.